<?php

namespace Xfrocks\AttachmentStore\Data;

use Aws\CloudFront\CloudFrontClient;
use XF\Db\AbstractAdapter;
use Xfrocks\AttachmentStore\DelayQueue;
use Xfrocks\AttachmentStore\Listener;
use Xfrocks\AttachmentStore\Fs;
use Xfrocks\AttachmentStore\Util\EngineOptions;
use Xfrocks\AttachmentStore\Util\MaintainFilename;
use Xfrocks\AttachmentStore\XF\Entity\AttachmentData;
use XF\Util\File as FileUtil;

class Storage
{
    const FILE_PATH_EXISTS = 0;
    const FILE_PATH_NAME = 1;
    const FILE_PATH_ORIGINAL_ABSTRACTED_PATH = 2;

    const LATEST_FILE_URLS_MAX_ENTRIES = 10;

    /**
     * @var string
     */
    private $externalDataPathPrefix = 'attachment-files/';

    /**
     * @var array
     */
    private $filePathProfiles = [];

    /**
     * @var array
     */
    private $latestFileUrls = [];

    /**
     * @param AbstractAdapter $db
     * @return DelayQueue
     */
    public function delayQueue($db = null)
    {
        static $delayQueue = null;

        if ($delayQueue === null) {
            $delayQueue = new DelayQueue($db ?: \XF::db());
        }

        return $delayQueue;
    }

    /**
     * @param AttachmentData $data
     * @param string $filePath
     * @param string $originalAbstractedPath
     * @return string
     */
    public function getAbstractedPathReplacement($data, $filePath, $originalAbstractedPath)
    {
        switch ($data->bdattachmentstore_engine) {
            case Listener::MODE_EXTERNAL_DATA:
                $abstractedPath = $this->getFilePathExternalData($filePath);
                break;
            case Listener::MODE_FTP:
            case Listener::MODE_GCS:
            case Listener::MODE_S3:
                $abstractedPath = Fs::buildPath(
                    $data->bdattachmentstore_engine,
                    $data->bdattachmentstore_options,
                    $filePath
                );
                break;
            default:
                throw new \InvalidArgumentException('Unsupported storage engine: ' . $data->bdattachmentstore_engine);
        }

        $this->filePathProfiles[$filePath][self::FILE_PATH_NAME] = $data->filename;
        $this->filePathProfiles[$filePath][self::FILE_PATH_ORIGINAL_ABSTRACTED_PATH] = $originalAbstractedPath;

        if (!isset($this->filePathProfiles[$filePath][self::FILE_PATH_EXISTS]) &&
            $data->getOption(AttachmentData::XFROCKS_ATTACHMENTSTORE_OPTION_IS_NEWLY_INSERTED) === true
        ) {
            $this->filePathProfiles[$filePath][self::FILE_PATH_EXISTS] = false;
        }

        return $abstractedPath;
    }

    /**
     * @param string $filePath
     * @param int|null $key
     * @return mixed|null
     */
    public function getFilePathProfile($filePath, $key = null)
    {
        if (!isset($this->filePathProfiles[$filePath])) {
            return null;
        }
        $profile = $this->filePathProfiles[$filePath];

        if ($key === null) {
            return $profile;
        }

        if (!isset($profile[$key])) {
            return null;
        }

        return $profile[$key];
    }

    /**
     * @param string|null $key
     * @return mixed|null
     */
    public function getSystemConfig($key = null)
    {
        static $config = null;

        if ($config === null) {
            $config = \XF::options()->bdAttachmentStore_config;
        }

        if ($key !== null) {
            if (isset($config[$key])) {
                return $config[$key];
            } else {
                return null;
            }
        }

        return $config;
    }

    /**
     * @return string
     */
    public function getDefaultEngine()
    {
        return strval($this->getSystemConfig('mode'));
    }

    /**
     * @param string $engine
     * @param AttachmentData $data
     * @return array
     */
    public function getDefaultOptions($engine, $data)
    {
        $options = $this->getDefaultOptionsPure($engine);

        $maintainFilename = boolval($this->getSystemConfig('maintainFilename'));
        if ($maintainFilename && strlen($data->filename) > 0) {
            $options['filename'] = MaintainFilename::getSafeFileName($data->filename);
            $options['filenameFormat'] = $this->getSystemConfig('maintainFilenameFormat');
        }

        return $options;
    }

    /**
     * @param string|false $engine
     * @return array
     */
    public function getDefaultOptionsPure($engine)
    {
        if ($engine === false) {
            return [];
        }

        switch ($engine) {
            case Listener::MODE_ATTACHMENT:
            case Listener::MODE_EXTERNAL_DATA:
                $options = [];
                break;
            case Listener::MODE_FTP:
            case Listener::MODE_GCS:
            case Listener::MODE_S3:
                $config = $this->getSystemConfig();
                $options = $config[$engine];
                break;
            default:
                throw new \InvalidArgumentException(sprintf('Unsupported storage engine: %s', $engine));
        }

        if (!is_array($options)) {
            return [];
        }

        $filteredOptions = array_filter($options, function ($optionValue) {
            return $optionValue !== '';
        });

        return $filteredOptions;
    }

    /**
     * @param AttachmentData $data
     * @return string
     */
    public function getFilePathData($data)
    {
        $ext = FileUtil::getFileExtension($data->filename);
        if ($ext === '') {
            $ext = 'dat';
        }

        $filePath = sprintf('%s/%d_%s.%s', $this->getFilePathDate($data), $data->data_id, $data->file_hash, $ext);

        if (isset($data->bdattachmentstore_options['filename'])) {
            $filePath = MaintainFilename::getFilePath($data->bdattachmentstore_options, $filePath);
        }

        return $filePath;
    }

    /**
     * @param AttachmentData $data
     * @return string
     */
    public function getFilePathThumbnail($data)
    {
        $filePath = sprintf('%s/%d_%s_thumb.jpg', $this->getFilePathDate($data), $data->data_id, $data->file_hash);

        if (isset($data->bdattachmentstore_options['filename'])) {
            $filePath = MaintainFilename::getFilePath($data->bdattachmentstore_options, $filePath);
        }

        return $filePath;
    }

    /**
     * @param AttachmentData $data
     * @param string $filePath
     * @return string
     */
    public function getFileUrlWithAttachmentData($data, $filePath)
    {
        $fileUrl = $this->getFileUrlWithEngineOptions(
            $data->bdattachmentstore_engine,
            $data->bdattachmentstore_options,
            $filePath
        );

        while (count($this->latestFileUrls) > self::LATEST_FILE_URLS_MAX_ENTRIES) {
            array_shift($this->latestFileUrls);
        }
        $this->latestFileUrls[] = $fileUrl;

        return $fileUrl;
    }

    /**
     * @param string $engine
     * @param array $engineOptions
     * @param string $filePath
     * @return string
     */
    public function getFileUrlWithEngineOptions($engine, $engineOptions, $filePath)
    {
        switch ($engine) {
            case Listener::MODE_EXTERNAL_DATA:
                return $this->getFileUrlExternalData($filePath);
            case Listener::MODE_FTP:
                return $this->getFileUrlFtp($engineOptions, $filePath);
            case Listener::MODE_GCS:
                return $this->getFileUrlGcs($engineOptions, $filePath);
            case Listener::MODE_S3:
                return $this->getFileUrlS3($engineOptions, $filePath);
            default:
                throw new \InvalidArgumentException(sprintf('Unsupported engine %s', $engine));
        }
    }

    /**
     * @param AttachmentData $data
     * @return bool
     */
    public function isIgnored($data)
    {
        if (strlen($data->file_path) > 0) {
            // attachment owner specified custom data file_path
            // most likely it will alter data after uploading -> ignore it
            // unless the path is whitelisted by admin
            // to whitelist all, set whitelist='*' in config.php (string of 1 character of asterisk)
            $whitelist = \XF::config(Listener::CONFIG_FILE_PATH_WHITELIST);
            if ($whitelist !== '*') {
                if (!is_array($whitelist)) {
                    $whitelist = [
                        // XF2 video attachment files:
                        'data://video/%FLOOR%/%DATA_ID%-%HASH%.mp4',
                    ];
                }

                if (!in_array($data->file_path, $whitelist, true)) {
                    return true;
                }
            }
        }

        static $ignoredExts = null;
        if ($ignoredExts === null) {
            $ignoredExts = preg_split('/\s+/', \XF::options()->bdAttachmentStore_ignoredExts);
            if ($ignoredExts === false) {
                $ignoredExts = [];
            }
            $ignoredExts = array_map('trim', $ignoredExts);
            $ignoredExts = array_map('strtolower', $ignoredExts);
        }
        if (count($ignoredExts) > 0
            && strlen($data->filename) > 0
            && in_array(FileUtil::getFileExtension($data->filename), $ignoredExts, true)
        ) {
            // extension ignored
            return true;
        }

        return false;
    }

    /**
     * @param string $url
     * @return bool
     */
    public function isOurUrl($url)
    {
        return in_array($url, $this->latestFileUrls, true);
    }

    /**
     * @return void
     */
    public function log()
    {
        if (!\XF::$debugMode) {
            return;
        }

        $args = func_get_args();
        $logEntry = call_user_func_array('sprintf', $args);
        FileUtil::log('Xfrocks_AttachmentStore', $logEntry);
    }

    /**
     * @param string $filePath
     * @param int $key
     * @param mixed $value
     * @return mixed
     */
    public function setFilePathProfile($filePath, $key, $value)
    {
        $this->filePathProfiles[$filePath][$key] = $value;

        return $this->filePathProfiles[$filePath];
    }

    /**
     * @param AttachmentData $data
     * @return bool
     */
    public function shouldBeDelayed($data)
    {
        static $optionValue = null;

        if ($optionValue === null) {
            $optionValue = \XF::app()->options()->bdAttachmentStore_delay > 0;
        }

        return $optionValue;
    }

    /**
     * @param AttachmentData $data
     * @return string
     */
    private function getFilePathDate($data)
    {
        return gmdate('Y/m', $data->upload_date);
    }

    /**
     * @param string $filePath
     * @return string
     */
    private function getFilePathExternalData($filePath)
    {
        $externalDataPath = 'data://' . $this->externalDataPathPrefix . $filePath;

        return $externalDataPath;
    }

    /**
     * @param string $filePath
     * @return string
     */
    private function getFileUrlExternalData($filePath)
    {
        return \XF::app()->applyExternalDataUrl($this->externalDataPathPrefix . $filePath, true);
    }

    /**
     * @param array $engineOptions
     * @param string $filePath
     * @return string
     */
    private function getFileUrlFtp($engineOptions, $filePath)
    {
        $ftpConfig = EngineOptions::getFtpConfig($engineOptions);
        return sprintf('%s/%s', rtrim($ftpConfig[EngineOptions::FTP_URL], '/'), $filePath);
    }

    /**
     * @param array $engineOptions
     * @param string $filePath
     * @return string
     */
    private function getFileUrlGcs($engineOptions, $filePath)
    {
        $gcsConfig = EngineOptions::getGcsConfig($engineOptions);

        if (isset($gcsConfig[EngineOptions::GCS_URL])) {
            return sprintf('%s/%s', rtrim($gcsConfig[EngineOptions::GCS_URL], '/'), $filePath);
        }

        return sprintf(
            'https://storage.googleapis.com/%s/%s',
            $gcsConfig[EngineOptions::GCS_BUCKET],
            $filePath
        );
    }

    /**
     * @param array $engineOptions
     * @param string $filePath
     * @return string
     */
    private function getFileUrlS3($engineOptions, $filePath)
    {
        $s3Config = EngineOptions::getS3Config($engineOptions);
        $cfConfig = $this->getSystemConfig(Listener::GROUP_CLOUDFRONT);

        if (isset($s3Config[EngineOptions::S3_URL])) {
            $s3Url = sprintf('%s/%s', rtrim($s3Config[EngineOptions::S3_URL], '/'), $filePath);
        } else {
            $protocol = $this->getS3Protocol($s3Config);
            if (isset($cfConfig[EngineOptions::CLOUDFRONT_DOMAIN_NAME]) && strlen($cfConfig[EngineOptions::CLOUDFRONT_DOMAIN_NAME]) > 0) {
                $s3Url = sprintf('%s://%s/%s', $protocol, $cfConfig[EngineOptions::CLOUDFRONT_DOMAIN_NAME], $filePath);
            } else {
                $bucket = $s3Config[EngineOptions::S3_BUCKET];
                if (strpos($bucket, '.') !== false) {
                    $s3Url = sprintf('%s://%s/%s', $protocol, $bucket, $filePath);
                } else {
                    $s3Url = sprintf('%s://%s.s3.amazonaws.com/%s', $protocol, $bucket, $filePath);
                }
            }
        }

        if (isset($cfConfig[EngineOptions::CLOUDFRONT_ENABLED]) && boolval($cfConfig[EngineOptions::CLOUDFRONT_ENABLED])) {
            $cloudFrontUrl = $this->getFileUrlS3CloudFront($cfConfig, $s3Config, $s3Url);
            if ($cloudFrontUrl !== null) {
                return $cloudFrontUrl;
            }
        }

        return $s3Url;
    }

    /**
     * @param array $cfConfig
     * @param array $s3Config
     * @param string $s3Url
     * @return string|null
     */
    private function getFileUrlS3CloudFront($cfConfig, $s3Config, $s3Url)
    {
        if (!isset($cfConfig[EngineOptions::CLOUDFRONT_ENABLED]) || !boolval($cfConfig[EngineOptions::CLOUDFRONT_ENABLED])) {
            return null;
        }

        if (!isset($cfConfig[EngineOptions::CLOUDFRONT_TTL])) {
            throw new \InvalidArgumentException('TTL for CloudFront is missing');
        }
        $expires = \XF::$time + $cfConfig[EngineOptions::CLOUDFRONT_TTL];

        if (!isset($cfConfig[EngineOptions::CLOUDFRONT_KEY_PAIR_ID])) {
            throw new \InvalidArgumentException('Key Pair ID for CloudFront is missing');
        }
        $keyPairId = $cfConfig[EngineOptions::CLOUDFRONT_KEY_PAIR_ID];

        if (isset($cfConfig[EngineOptions::CLOUDFRONT_PK_FILE_PATH])) {
            $pkFilePath = $cfConfig[EngineOptions::CLOUDFRONT_PK_FILE_PATH];
        } elseif (isset($cfConfig[EngineOptions::CLOUDFRONT_PK])) {
            // legacy support
            $pkFilePath = FileUtil::getTempFile();
            file_put_contents($pkFilePath, $cfConfig[EngineOptions::CLOUDFRONT_PK]);
        } else {
            throw new \InvalidArgumentException('Private key for CloudFront is missing');
        }

        static $client = null;
        if ($client === null) {
            $client = new CloudFrontClient([
                'region' => 'ap-south-1',
                'version' => '2017-03-25',
            ]);
        }

        return $client->getSignedUrl([
            'url' => $s3Url,
            'expires' => $expires,
            'key_pair_id' => $keyPairId,
            'private_key' => $pkFilePath
        ]);
    }

    /**
     * @param mixed|array $s3Config
     * @return string
     */
    private function getS3Protocol($s3Config)
    {
        if (is_array($s3Config)
            && isset($s3Config[EngineOptions::S3_PROTOCOL])
        ) {
            $protocol = $s3Config[EngineOptions::S3_PROTOCOL];
            switch ($protocol) {
                case EngineOptions::S3_PROTOCOL_HTTP:
                case EngineOptions::S3_PROTOCOL_HTTPS:
                    return $protocol;
            }
        }

        return \XF::app()->request()->getProtocol();
    }
}
