<?php

namespace Xfrocks\AttachmentStore;

use XF\App;
use XF\Db\AbstractAdapter;
use Xfrocks\AttachmentStore\Data\Storage;
use Xfrocks\AttachmentStore\Service\Mover;
use Xfrocks\AttachmentStore\XF\Entity\AttachmentData;

class DelayQueue
{
    const JOB_KEY = 'Xfrocks_AttachmentStore_DelayQueue';

    /**
     * @var AbstractAdapter
     */
    protected $db;

    /**
     * @param AbstractAdapter $db
     */
    public function __construct($db)
    {
        $this->db = $db;
    }

    /**
     * @param AttachmentData $data
     * @return bool
     */
    public function queue($data)
    {
        $this->db->insert('xf_bdattachmentstore_delay', [
            'data_id' => $data->data_id,
            'insert_date' => time()
        ]);

        $jobManager = \XF::app()->jobManager();
        if (!$jobManager->getUniqueJob(self::JOB_KEY)) {
            try {
                $jobManager->enqueueUnique(self::JOB_KEY, 'Xfrocks\AttachmentStore\Job\DelayQueue', [], false);
            } catch (\Exception $e) {
                // just ignore it, this queue does best effort delivery
            }
        }

        return true;
    }

    /**
     * @param App $app
     * @param int $maxRunTime
     * @return void
     */
    public function run($app, $maxRunTime)
    {
        $s = microtime(true);
        $db = $this->db;

        /** @var Storage $storage */
        $storage = $app->data('Xfrocks\AttachmentStore:Storage');
        $defaultEngine = $storage->getDefaultEngine();
        /** @var Mover $mover */
        $mover = $app->service('Xfrocks\AttachmentStore:Mover');

        do {
            $queue = $this->getQueue();
            if (count($queue) === 0) {
                return;
            }
            $records = \XF::app()->em()->findByIds('XF:AttachmentData', array_keys($queue));

            foreach ($queue as $dataId => $record) {
                if (!$db->delete('xf_bdattachmentstore_delay', 'data_id = ?', $dataId)) {
                    // already been deleted - run elsewhere
                    continue;
                }

                if (!isset($records[$dataId])) {
                    continue;
                }
                /** @var AttachmentData $data */
                $data = $records[$dataId];

                if ($storage->isIgnored($data)) {
                    continue;
                }

                try {
                    $startTime = microtime(true);
                    $newData = $mover->move($data, $defaultEngine);
                    if ($newData !== null) {
                        $elapsed = microtime(true) - $startTime;
                        $storage->log(sprintf('Queue #%d -> #%d in %f', $data->data_id, $newData->data_id, $elapsed));
                    }
                } catch (\Exception $e) {
                    \XF::logException($e, false, sprintf('Could not move #%d: ', $data->data_id));
                }

                if ($maxRunTime > 0 && microtime(true) - $s > $maxRunTime) {
                    return;
                }
            }
        } while ($queue);
    }

    /**
     * @param int $limit
     * @return array
     */
    public function getQueue($limit = 20)
    {
        $db = $this->db;

        return $db->fetchAllKeyed($db->limit('
            SELECT *
            FROM xf_bdattachmentstore_delay
            ORDER BY insert_date
        ', $limit), 'data_id');
    }

    /**
     * @return bool
     */
    public function hasMore()
    {
        return (bool)$this->db->fetchOne('
            SELECT MIN(data_id)
            FROM xf_bdattachmentstore_delay
        ');
    }
}
