<?php

namespace Xfrocks\AttachmentStore;

use Aws\S3\S3Client;
use Google\Cloud\Storage\StorageClient as GcsClient;
use League\Flysystem\Adapter\Ftp;
use League\Flysystem\AdapterInterface;
use League\Flysystem\AwsS3v3\AwsS3Adapter;
use League\Flysystem\EventableFilesystem\EventableFilesystem;
use \XF\Util\File as FileUtil;
use Xfrocks\AttachmentStore\Data\Storage;
use Xfrocks\AttachmentStore\FsAdapter\GcsAdapter;
use Xfrocks\AttachmentStore\Util\EngineOptions;
use Xfrocks\AttachmentStore\Util\MaintainFilename;
use Xfrocks\AttachmentStore\Util\MimeType;

class Fs extends EventableFilesystem
{
    /**
     * @var bool
     */
    private $keepLocalCopy = false;

    public function __construct(AdapterInterface $adapter, array $config)
    {
        parent::__construct($adapter, $config);

        $this->keepLocalCopy = $this->storage()->getSystemConfig('keepLocalCopy');
    }

    /**
     * @param string $method
     * @param array $arguments
     * @return mixed
     * @throws \Exception
     */
    public function delegateMethodCall($method, array $arguments = [])
    {
        if (!\XF::$debugMode) {
            return parent::delegateMethodCall($method, $arguments);
        }

        $strings = [];
        foreach ($arguments as $argument) {
            if (is_string($argument)) {
                $strings[] = $argument;
            }
        }

        try {
            $startTime = microtime(true);

            $result = parent::delegateMethodCall($method, $arguments);

            $resultString = '';
            if (is_bool($result)) {
                $resultString = sprintf(' = %s', $result ? 'true' : 'false');
            }

            $elapsed = microtime(true) - $startTime;
            $this->log('%s(%s)%s in %fs', $method, implode(', ', $strings), $resultString, $elapsed);

            return $result;
        } catch (\Exception $e) {
            $this->log('%s(%s) with exception %s', $method, implode(', ', $strings), $e->getMessage());

            throw $e;
        }
    }

    /**
     * @param string $path
     * @param array $config
     * @return bool
     * @throws \Exception
     */
    public function delete($path, array $config = [])
    {
        $result = false;
        $exception = null;
        try {
            $result = parent::delete($path, $config);
        } catch (\Exception $e) {
            $exception = $e;
        }

        if ($this->keepLocalCopy) {
            $oap = $this->storage()->getFilePathProfile($path, Storage::FILE_PATH_ORIGINAL_ABSTRACTED_PATH);
            if (strlen($oap) > 0) {
                FileUtil::deleteFromAbstractedPath($oap);
                $this->log('delete(%s) no $originalAbstractedPath', $path);
            } else {
                $this->log('delete(%s) also deleted %s', $path, $oap);
            }
        }

        if ($result) {
            $this->storage()->setFilePathProfile($path, Storage::FILE_PATH_EXISTS, false);
        }

        if ($exception !== null) {
            throw $exception;
        }

        return $result;
    }

    public function has($path, array $config = [])
    {
        $filePathProfileExists = $this->storage()->getFilePathProfile($path, Storage::FILE_PATH_EXISTS);
        if (is_bool($filePathProfileExists)) {
            $this->log('has(%s) = %d', $path, $filePathProfileExists ? 'true' : 'false');
            return $filePathProfileExists;
        }

        return parent::has($path, $config);
    }

    /**
     * @param string $path
     * @param resource $resource
     * @param array $config
     * @return bool
     * @throws \Exception
     */
    public function putStream($path, $resource, array $config = [])
    {
        $storage = $this->storage();

        // set additional metadata for S3
        $originalFileName = $storage->getFilePathProfile($path, Storage::FILE_PATH_NAME);
        if (strlen($originalFileName) > 0) {
            $config['ContentType'] = MimeType::guessByFileName($originalFileName);

            $safeFileName = MaintainFilename::getSafeFileName($originalFileName);
            $config['ContentDisposition'] = sprintf('attachment; filename=%s;', $safeFileName);
        }

        $oap = '';
        $needsTempFile = false;
        if ($this->keepLocalCopy) {
            $filePathProfileOap = $storage->getFilePathProfile($path, Storage::FILE_PATH_ORIGINAL_ABSTRACTED_PATH);
            if (is_string($filePathProfileOap) && strlen($filePathProfileOap) > 0) {
                $oap = $filePathProfileOap;
                $needsTempFile = true;
            }
        }

        $tempFile = '';
        if ($needsTempFile) {
            $tempFile = FileUtil::copyStreamToTempFile($resource);
            if ($tempFile === false) {
                $storage->log('putStream(%s) tempFile === false', $path);
                return false;
            }
            $storage->log('putStream(%s) tempFile %s', $path, $tempFile);
        }

        $result = false;
        $exception = null;
        try {
            $fp = $resource;
            $needsFclose = false;
            if ($tempFile !== '') {
                /** @var resource $fp */
                $fp = fopen($tempFile, 'r');
                $needsFclose = true;
            }

            $result = parent::putStream($path, $fp, $config);

            /** @var resource|false $fpAfter */
            $fpAfter = $fp;
            if ($needsFclose && is_resource($fpAfter)) {
                // the adapter may close the resource by itself, pre-checking is required
                fclose($fpAfter);
            }
        } catch (\Exception $e) {
            $exception = $e;
        }

        if ($result === true && $oap !== '' && $tempFile !== '') {
            try {
                $result = FileUtil::copyFileToAbstractedPath($tempFile, $oap);
                $storage->log('putStream(%s) keepLocalCopy at %s', $path, $oap);
            } catch (\Exception $e) {
                $this->log('putStream(%s) ignored exception while keepLocalCopy %s', $e->getMessage());
            }
        }

        if ($result === true) {
            $storage->setFilePathProfile($path, Storage::FILE_PATH_EXISTS, true);
        }

        if ($exception !== null) {
            throw $exception;
        }

        return $result;
    }

    /**
     * @param string $prefix
     * @return void
     */
    private function setup($prefix)
    {
        \XF::app()->fs()->mountFilesystem($prefix, $this);

        $this->log('setup(%s) OK', $prefix);
    }

    /**
     * @return void
     */
    private function log()
    {
        if (!\XF::$debugMode) {
            return;
        }

        $args = func_get_args();
        if (count($args) > 0 && is_string($args[0])) {
            $adapter = get_class($this->adapter);
            $adapter = preg_replace('/^.+\\\\/', '', $adapter);
            $args[0] = sprintf('[%s] %s', $adapter, $args[0]);
        }

        call_user_func_array([$this->storage(), 'log'], $args);
    }

    /**
     * @return Storage
     */
    private function storage()
    {
        return \XF::app()->data('Xfrocks\AttachmentStore:Storage');
    }

    /**
     * @param string $engine
     * @param mixed $storageOptions
     * @param string $path
     * @return string
     */
    public static function buildPath($engine, $storageOptions, $path)
    {
        $prefix = self::registerIfNeeded($engine, is_array($storageOptions) ? $storageOptions : []);
        return sprintf('%s://%s', $prefix, $path);
    }

    /**
     * @param string $engine
     * @param array $engineOptions
     * @return array
     */
    protected static function getAdapterConfig($engine, array $engineOptions)
    {
        switch ($engine) {
            case Listener::MODE_FTP:
                $ftp = EngineOptions::getFtpConfig($engineOptions);
                $adapter = [
                    'fs' => [
                        'visibility' => AdapterInterface::VISIBILITY_PUBLIC,
                    ],
                    'ftp' => [
                        'host' => $ftp[EngineOptions::FTP_HOST],
                        'port' => $ftp[EngineOptions::FTP_PORT] ?: 21,
                        'passive' => $ftp[EngineOptions::FTP_PASSIVE] === true,
                    ]
                ];

                if (strlen($ftp[EngineOptions::FTP_USERNAME]) > 0 && strlen($ftp[EngineOptions::FTP_PASSWORD]) > 0) {
                    $adapter['ftp']['username'] = $ftp[EngineOptions::FTP_USERNAME];
                    $adapter['ftp']['password'] = $ftp[EngineOptions::FTP_PASSWORD];
                }
                $adapter['ftp']['root'] = $ftp[EngineOptions::FTP_PATH];

                if ($ftp[EngineOptions::FTP_CALLBACK] !== null) {
                    $adapter = $ftp[EngineOptions::FTP_CALLBACK]($adapter);
                }

                return $adapter;
            case Listener::MODE_GCS:
                $gcs = EngineOptions::getGcsConfig($engineOptions);
                $adapter = [
                    'fs' => [
                        'visibility' => AdapterInterface::VISIBILITY_PUBLIC,
                    ],
                    'gcsAdapter' => [
                        'bucket' => $gcs[EngineOptions::GCS_BUCKET],
                    ],
                    'gcsClient' => [
                        'keyFilePath' => $gcs[EngineOptions::GCS_KEY_FILE_PATH],
                        'projectId' => $gcs[EngineOptions::GCS_PROJECT_ID],
                    ]
                ];

                if ($gcs[EngineOptions::GCS_CALLBACK] !== null) {
                    $adapter = $gcs[EngineOptions::GCS_CALLBACK]($adapter);
                }

                return $adapter;
            case Listener::MODE_S3:
                $s3 = EngineOptions::getS3Config($engineOptions);
                $adapter = [
                    'bucket' => $s3[EngineOptions::S3_BUCKET],
                    'fs' => [
                        'visibility' => ($s3[EngineOptions::S3_PRIVATE] === true) ?
                            AdapterInterface::VISIBILITY_PRIVATE :
                            AdapterInterface::VISIBILITY_PUBLIC,
                    ],
                    's3Client' => [
                        'credentials' => [
                            'key' => $s3[EngineOptions::S3_KEY],
                            'secret' => $s3[EngineOptions::S3_SECRET],
                        ],
                        'region' => $s3[EngineOptions::S3_REGION] ?: 'us-east-1',
                        'version' => '2006-03-01',
                    ]
                ];

                if (strlen($s3[EngineOptions::S3_ENDPOINT]) > 0) {
                    $adapter['s3Client']['endpoint'] = $s3[EngineOptions::S3_ENDPOINT];

                    if (strpos($s3[EngineOptions::S3_ENDPOINT], $s3[EngineOptions::S3_BUCKET]) !== false) {
                        $adapter['s3Client']['bucket_endpoint'] = true;
                    } else {
                        $adapter['s3Client']['use_path_style_endpoint'] = true;
                    }
                }

                if ($s3[EngineOptions::S3_CALLBACK] !== null) {
                    $adapter = $s3[EngineOptions::S3_CALLBACK]($adapter);
                }

                return $adapter;
        }

        throw new \InvalidArgumentException(sprintf('Unsupported engine %s', $engine));
    }

    /**
     * @param string $engine
     * @param array $engineOptions
     * @return AdapterInterface
     */
    public static function getAdapterCustom($engine, array $engineOptions)
    {
        $config = self::getAdapterConfig($engine, $engineOptions);
        return self::getAdapterObj($engine, $config);
    }

    /**
     * @return AdapterInterface|null
     */
    public static function getAdapterDefault()
    {
        /** @var Storage $storage */
        $storage = \XF::app()->data('Xfrocks\AttachmentStore:Storage');

        $engine = $storage->getDefaultEngine();
        if ($engine === '') {
            return null;
        }

        $engineOptions = $storage->getDefaultOptionsPure($engine);
        $config = self::getAdapterConfig($engine, $engineOptions);

        return self::getAdapterObj($engine, $config);
    }

    /**
     * @param string $engine
     * @param array $config
     * @return AdapterInterface
     */
    protected static function getAdapterObj($engine, array $config)
    {
        switch ($engine) {
            case Listener::MODE_FTP:
                return new Ftp($config['ftp']);
            case Listener::MODE_GCS:
                $gcsClient = new GcsClient($config['gcsClient']);
                return new GcsAdapter($gcsClient, $config['gcsAdapter']);
            case Listener::MODE_S3:
                $s3Client = new S3Client($config['s3Client']);
                return new AwsS3Adapter($s3Client, $config['bucket']);
        }

        throw new \InvalidArgumentException(sprintf('Unsupported engine %s', $engine));
    }

    /**
     * @param string $engine
     * @param array $engineOptions
     * @return string
     */
    public static function registerIfNeeded($engine, array $engineOptions)
    {
        static $prefixes = [];

        $config = self::getAdapterConfig($engine, $engineOptions);
        $json = json_encode($config);
        if ($json === false) {
            throw new \RuntimeException(sprintf('json_encode(%s) returned false', var_export($config, true)));
        }

        $hash = $engine . md5($json);
        if (isset($prefixes[$hash])) {
            return $prefixes[$hash];
        }

        $prefix = sprintf('%s-%s%d', Listener::STREAM_WRAPPER_PROTOCOL, $engine, count($prefixes));
        $adapter = self::getAdapterObj($engine, $config);

        $fs = new self($adapter, $config['fs']);
        $fs->setup($prefix);

        $prefixes[$hash] = $prefix;

        return $prefixes[$hash];
    }
}
