<?php

namespace Xfrocks\AttachmentStore\Job;

use XF\Job\AbstractJob;
use Xfrocks\AttachmentStore\Data\Storage;
use Xfrocks\AttachmentStore\Listener;
use Xfrocks\AttachmentStore\Service\Mover;
use Xfrocks\AttachmentStore\XF\Entity\AttachmentData;

class Move extends AbstractJob
{
    /**
     * @var array
     */
    protected $defaultData = [
        'start' => 0,
        'batch' => 100,
        'backupPath' => '',

        'isCli' => false,
        'modulo' => '',
    ];

    /**
     * @var Storage
     */
    private $storage;

    /**
     * @param \XF\App $app
     * @param mixed $jobId
     * @param array $data
     */
    public function __construct(\XF\App $app, $jobId, array $data = [])
    {
        parent::__construct($app, $jobId, $data);

        $this->storage = $this->app->data('Xfrocks\AttachmentStore:Storage');
    }

    public function run($maxRunTime)
    {
        $startTime = microtime(true);

        $db = $this->app->db();
        $defaultEngine = $this->storage->getDefaultEngine();
        /** @var Mover $mover */
        $mover = $this->app->service('Xfrocks\AttachmentStore:Mover');

        if ($this->data['backupPath'] !== '') {
            if (!isset($this->data['preparedBackupPath'])) {
                $this->data['preparedBackupPath'] = $mover->backupPathPrepare($this->data['backupPath']);
                return $this->resume();
            } else {
                $mover->backupPathUsePrepared($this->data['preparedBackupPath']);
            }
        }

        if ($this->data['modulo'] !== '') {
            $modulo = $this->data['modulo'];
            $this->data['modulo'] = '';
            unset($this->data['preparedModulo']);

            $moduloParts = preg_split('/[^0-9]/', $modulo, -1, PREG_SPLIT_NO_EMPTY);
            if (!is_array($moduloParts) || count($moduloParts) !== 2) {
                $this->log('modulo must be in the format "n,m" (n comma m)');
                return $this->complete();
            }

            $this->data['preparedModulo'] = array_map('intval', $moduloParts);
        }

        $dataIds = $db->fetchAllColumn($db->limit('
            SELECT data_id
            FROM xf_attachment_data
            WHERE data_id > ? AND attach_count > 0
            ORDER BY data_id
        ', $this->data['batch']), $this->data['start']);
        if (!$dataIds) {
            $this->log('Done');
            return $this->complete();
        }

        $records = $this->app->em()->findByIds('XF:AttachmentData', $dataIds);

        $done = 0;

        foreach ($dataIds as $dataId) {
            $this->data['start'] = $dataId;

            if (isset($this->data['preparedModulo'])) {
                $modulo = $dataId % $this->data['preparedModulo'][0];
                if ($modulo !== $this->data['preparedModulo'][1]) {
                    $this->log(sprintf('Skipped #%d (modulo=%d)', $dataId, $modulo));
                    continue;
                }
            }

            if (!isset($records[$dataId])) {
                continue;
            }
            /** @var AttachmentData $data */
            $data = $records[$dataId];

            $neededEngine = $this->storage->isIgnored($data) ? Listener::MODE_ATTACHMENT : $defaultEngine;

            try {
                $startTime = microtime(true);
                $this->log(sprintf('Moving #%d at %f', $data->data_id, $startTime));

                $newData = $mover->move($data, $neededEngine);
                if ($newData !== null) {
                    $elapsed = microtime(true) - $startTime;
                    $this->log(sprintf('Moved #%d -> #%d in %f', $data->data_id, $newData->data_id, $elapsed));
                }
            } catch (\Exception $e) {
                \XF::logException($e, false, sprintf('Could not move #%d', $data->data_id));
                $this->log(sprintf('Could not move #%d: %s', $data->data_id, $e->getMessage()));
            }

            $done++;

            if (microtime(true) - $startTime >= $maxRunTime) {
                break;
            }
        }

        $this->data['batch'] = $this->calculateOptimalBatch($this->data['batch'], $done, $startTime, $maxRunTime, 1000);

        return $this->resume();
    }

    /**
     * @return string
     */
    public function getStatusMessage()
    {
        $actionPhrase = \XF::phrase('bdattachmentstore_moving_files');
        $typePhrase = \XF::phrase('attachments');
        return sprintf('%s... %s (%s)', $actionPhrase, $typePhrase, $this->data['start']);
    }

    /**
     * @return bool
     */
    public function canCancel()
    {
        return true;
    }

    /**
     * @return bool
     */
    public function canTriggerByChoice()
    {
        return true;
    }

    /**
     * @param string $message
     * @return void
     */
    private function log($message)
    {
        if ($this->data['isCli'] === true) {
            $this->storage->log($message);
        }

        echo("$message\n");
    }
}
