<?php

namespace Xfrocks\AttachmentStore\Job;

use XF\Job\AbstractJob;
use Xfrocks\AttachmentStore\Data\Storage;
use Xfrocks\AttachmentStore\Service\Updater;
use Xfrocks\AttachmentStore\XF\Entity\AttachmentData;

class Update extends AbstractJob
{
    /**
     * @var array
     */
    protected $defaultData = [
        'start' => 0,
        'batch' => 100,
        'backupTable' => '',
        'verify' => false,
        'rules' => []
    ];

    public function run($maxRunTime)
    {
        $startTime = microtime(true);

        $db = $this->app->db();
        /** @var Storage $storage */
        $storage = $this->app->data('Xfrocks\AttachmentStore:Storage');
        /** @var Updater $updater */
        $updater = $this->app->service('Xfrocks\AttachmentStore:Updater');

        if (!$this->data['backupTable']) {
            $this->data['backupTable'] = $updater->createBackupTable();
        }

        $dataIds = $db->fetchAllColumn($db->limit('
            SELECT data_id
            FROM xf_attachment_data
            WHERE data_id > ? AND attach_count > 0
            ORDER BY data_id
        ', $this->data['batch']), $this->data['start']);
        if (!$dataIds) {
            return $this->complete();
        }

        $records = $this->app->em()->findByIds('XF:AttachmentData', $dataIds);

        $done = 0;

        foreach ($dataIds as $dataId) {
            $this->data['start'] = $dataId;

            if (!isset($records[$dataId])) {
                continue;
            }
            /** @var AttachmentData $data */
            $data = $records[$dataId];

            try {
                $startTime = microtime(true);

                if ($updater->checkRules($data, $this->data['rules'])) {
                    if ($updater->update($data, $this->data['backupTable'], $this->data['verify'])) {
                        $elapsed = microtime(true) - $startTime;
                        $storage->log(sprintf('Updated #%d in %f', $data->data_id, $elapsed));
                    }
                }
            } catch (\Exception $e) {
                \XF::logException($e, false, sprintf('Could not update #%d:', $data->data_id));
            }

            $done++;

            if (microtime(true) - $startTime >= $maxRunTime) {
                break;
            }
        }

        $this->data['batch'] = $this->calculateOptimalBatch($this->data['batch'], $done, $startTime, $maxRunTime, 1000);

        return $this->resume();
    }

    /**
     * @return string
     */
    public function getStatusMessage()
    {
        $actionPhrase = \XF::phrase('bdattachmentstore_updating_storage_options');
        $typePhrase = \XF::phrase('attachments');
        return sprintf('%s... %s (%s)', $actionPhrase, $typePhrase, $this->data['start']);
    }

    /**
     * @return bool
     */
    public function canCancel()
    {
        return true;
    }

    /**
     * @return bool
     */
    public function canTriggerByChoice()
    {
        return true;
    }
}
