<?php

namespace Xfrocks\AttachmentStore;

use XF\Mvc\RouteBuiltLink;
use Xfrocks\AttachmentStore\Data\Storage;
use Xfrocks\AttachmentStore\Util\EngineOptions;
use Xfrocks\AttachmentStore\Util\S3;
use Xfrocks\AttachmentStore\XF\Entity\AttachmentData;

class Listener
{
    const CONFIG_FILE_PATH_WHITELIST = 'bdAttachmentStore_filePathWhitelist';
    const CONFIG_FTP_SERVERS = 'bdAttachmentStore_ftpServers';
    const CONFIG_GCS_BUCKETS = 'bdAttachmentStore_gcsBuckets';
    const CONFIG_MIME_TYPES = 'bdAttachmentStore_mimeTypes';
    const CONFIG_S3_BUCKETS = 'bdAttachmentStore_s3Buckets';

    const MODE = 'mode';
    const MODE_ATTACHMENT = '';
    const MODE_EXTERNAL_DATA = 'external_data';
    const MODE_FTP = 'ftp';
    const MODE_GCS = 'gcs';
    const MODE_S3 = 's3';
    const GROUP_CLOUDFRONT = 'cloudFront';

    const STREAM_WRAPPER_PROTOCOL = 'Xfrocks-AttachmentStore';

    const ROUTE_ATTACHMENTS_PREFIX = 'attachments';
    const ROUTE_ATTACHMENTS_SUB_SECTION = '';

    /**
     * @var callable|null
     */
    private static $routeAttachmentsBuildCallbackOriginal = null;

    /**
     * @param \XF\Admin\App $app
     * @return void
     * @see S3::getRegions()
     */
    public static function appAdminSetup($app)
    {
        $app->container()->extend('templater', function ($templater) {
            /** @var \XF\Template\Templater $templater */
            $templater->addFunction(
                strtolower('Xfrocks_AttachmentStore_getS3Regions'),
                ['Xfrocks\AttachmentStore\Util\S3', 'getRegions']
            );

            return $templater;
        });
    }

    /**
     * @param \XF\Container $container
     * @param \XF\Mvc\Router $router
     * @return void
     */
    public static function routerPublicSetup($container, &$router)
    {
        $routes = $router->getRoutes();

        if (!isset($routes[self::ROUTE_ATTACHMENTS_PREFIX])) {
            return;
        }
        $routePrefix = $routes[self::ROUTE_ATTACHMENTS_PREFIX];

        if (!isset($routePrefix[self::ROUTE_ATTACHMENTS_SUB_SECTION])) {
            return;
        }
        $route = $routePrefix[self::ROUTE_ATTACHMENTS_SUB_SECTION];

        self::$routeAttachmentsBuildCallbackOriginal = $route['build_callback'];
        $route['build_callback'] = [__CLASS__, 'routeAttachmentsBuildCallback'];

        $router->addRoute(self::ROUTE_ATTACHMENTS_PREFIX, self::ROUTE_ATTACHMENTS_SUB_SECTION, $route);
    }

    /**
     * @param string $prefix
     * @param array $route
     * @param string $action
     * @param mixed $linkData
     * @param array $parameters
     * @param \XF\Mvc\Router $router
     * @return mixed
     */
    public static function routeAttachmentsBuildCallback(&$prefix, &$route, &$action, &$linkData, &$parameters, $router)
    {
        if (is_object($linkData) && $linkData instanceof \XF\Entity\Attachment) {
            $renderFileUrl = true;
            /** @var AttachmentData $attachmentData */
            $attachmentData = $linkData->Data;
            /** @var Storage $storage */
            $storage = \XF::app()->data('Xfrocks\AttachmentStore:Storage');

            if (strlen($attachmentData->bdattachmentstore_engine) === 0) {
                $renderFileUrl = false;
            }

            if ($renderFileUrl && $attachmentData->bdattachmentstore_engine == Listener::MODE_S3) {
                $cfConfig = $storage->getSystemConfig(self::GROUP_CLOUDFRONT);
                if (is_array($cfConfig) &&
                    isset($cfConfig[EngineOptions::CLOUDFRONT_ENABLED]) &&
                    $cfConfig[EngineOptions::CLOUDFRONT_ENABLED] === true
                ) {
                    // special case: disable our URLs for Amazon S3+CloudFront
                    // they will be taken care of by bdAttachmentStore_XenForo_ControllerPublic_Attachment
                    $renderFileUrl = false;
                }
            }

            if ($renderFileUrl && boolval($storage->getSystemConfig('checkPermission'))) {
                // the system has been configured to enforce permission
                // do not rewrite attachment links
                // since 0.10
                $renderFileUrl = false;
            }

            if ($renderFileUrl) {
                $filePath = $storage->getFilePathData($attachmentData);
                $fileUrl = $storage->getFileUrlWithAttachmentData($attachmentData, $filePath);
                $link = new RouteBuiltLink($fileUrl, false);
                return $link;
            }
        }

        if (self::$routeAttachmentsBuildCallbackOriginal === null) {
            return null;
        }

        $args = [&$prefix, &$route, &$action, &$linkData, &$parameters, $router];
        return call_user_func_array(self::$routeAttachmentsBuildCallbackOriginal, $args);
    }
}
