<?php

namespace Xfrocks\AttachmentStore\Service;

use League\Flysystem\MountManager;
use XF\Service\AbstractService;
use XF\Util\File;
use Xfrocks\AttachmentStore\Data\Storage;
use Xfrocks\AttachmentStore\Util\EngineOptions;
use Xfrocks\AttachmentStore\XF\Entity\AttachmentData;

use XF\Util\File as FileUtil;

class Mover extends AbstractService
{
    /**
     * @var string|null
     */
    protected $backupPathPrepared = null;

    /**
     * @param string $dirPath
     * @return string
     */
    public function backupPathPrepare($dirPath)
    {
        $listFilePath = $dirPath . DIRECTORY_SEPARATOR . 'list.txt';
        if (file_exists($listFilePath)) {
            return $listFilePath;
        }

        $dir = File::getTempDir();
        $tempFile = tempnam($dir, 'attachMover' . gmdate('Ymd_'));
        if ($tempFile === false) {
            throw new \RuntimeException('tempnam() returned false');
        }

        $cmd = sprintf('find %s -type f > %s', escapeshellarg($dirPath), escapeshellarg($tempFile));
        exec($cmd, $output, $returnVar);

        if ($returnVar !== 0) {
            throw new \RuntimeException($cmd . implode("\n", $output), $returnVar);
        }

        return $tempFile;
    }

    /**
     * @param string $filePath
     * @return void
     */
    public function backupPathUsePrepared($filePath)
    {
        $this->backupPathPrepared = $filePath;
    }

    /**
     * @param AttachmentData $data
     * @param string $neededEngine
     * @return AttachmentData|null
     * @throws \Exception
     */
    public function move($data, $neededEngine)
    {
        $neededOptions = $this->storage()->getDefaultOptions($neededEngine, $data);
        if (EngineOptions::checkStorageOptionsSame(
            $data->bdattachmentstore_engine,
            $data->bdattachmentstore_options,
            $neededEngine,
            $neededOptions
        )) {
            return null;
        }

        if (strlen($data->bdattachmentstore_engine) === 0 && !$data->isDataAvailable()) {
            throw new \Exception('Data is unavailable');
        }

        $db = $this->db();
        $em = $this->app->em();
        $fs = $this->app->fs();
        $dataArray = $data->toArray();
        $dataPath = $data->getAbstractedDataPath();
        $thumbnailPath = false;
        if ($data->thumbnail_width > 0 && $data->thumbnail_height > 0) {
            $thumbnailPath = $data->getAbstractedThumbnailPath();
        }

        $buggyDataPath = false;
        $buggyThumbnailPath = false;
        if (defined('XFROCKS_ATTACHMENT_STORE_20171220_MOVE_BUGGY_DATA') && strlen($data->bdattachmentstore_engine) > 0) {
            $buggyArray = $dataArray;
            unset($buggyArray['data_id']);
            /** @var AttachmentData $buggyData */
            $buggyData = $em->create('XF:AttachmentData');
            $buggyData->bulkSet($buggyArray, ['skipInvalid' => true]);
            $buggyData->setOption(AttachmentData::XFROCKS_ATTACHMENTSTORE_OPTION_IS_NEWLY_INSERTED, true);
            $buggyDataPath = $buggyData->getAbstractedDataPath();
            if ($thumbnailPath !== false) {
                $buggyThumbnailPath = $buggyData->getAbstractedThumbnailPath();
            }
        }

        $newArray = $dataArray;
        unset($newArray['data_id']);
        unset($newArray['bdattachmentstore_engine']);
        unset($newArray['bdattachmentstore_options']);

        /** @var AttachmentData $newData */
        $newData = $em->create('XF:AttachmentData');
        $newData->bulkSet($newArray, ['skipInvalid' => true]);
        if (strlen($neededEngine) > 0) {
            $newData->set('bdattachmentstore_engine', $neededEngine);
            $newData->set('bdattachmentstore_options', $neededOptions);
        }

        $db->beginTransaction();
        $newDataPath = null;
        $newThumbnailPath = null;

        try {
            $newData->save(true, false);
            $db->update('xf_attachment_data', ['attach_count' => 0], 'data_id = ?', $data->data_id);
            $db->update('xf_attachment', ['data_id' => $newData->data_id], 'data_id = ?', $data->data_id);

            $newDataPath = $newData->getAbstractedDataPath();
            $this->copy($fs, $dataPath, $newDataPath, $buggyDataPath, $data);

            if (is_string($thumbnailPath) && strlen($thumbnailPath) > 0) {
                $newThumbnailPath = $newData->getAbstractedThumbnailPath();
                try {
                    $this->copy($fs, $thumbnailPath, $newThumbnailPath, $buggyThumbnailPath);
                } catch (\Exception $e) {
                    // ignore thumbnail copy issues
                }
            }
        } catch (\Exception $e) {
            $em->detachEntity($newData);
            $db->rollback();

            if ($newDataPath !== null) {
                FileUtil::deleteFromAbstractedPath($newDataPath);
            }

            if ($newThumbnailPath !== null) {
                FileUtil::deleteFromAbstractedPath($newThumbnailPath);
            }

            throw $e;
        }

        $db->commit();

        return $newData;
    }

    /**
     * @return Storage
     */
    public function storage()
    {
        return $this->app->data('Xfrocks\AttachmentStore:Storage');
    }

    /**
     * @param MountManager $fs
     * @param string $path
     * @param string $newPath
     * @param string|false $buggyPath
     * @param AttachmentData|null $dataForBackupSearch
     * @return bool
     * @throws \Exception
     */
    protected function copy($fs, $path, $newPath, $buggyPath, $dataForBackupSearch = null)
    {
        $throwable = null;
        try {
            return $fs->copy($path, $newPath);
        } catch (\Exception $e) {
            $throwable = $e;
        }

        if (is_string($buggyPath) && strlen($buggyPath) > 0) {
            try {
                $fs->copy($buggyPath, $newPath);
            } catch (\Exception $e) {
                // ignore buggy copy issues
            }
        }

        if ($dataForBackupSearch !== null) {
            $fileSize = $dataForBackupSearch->file_size;
            $fileHash = $dataForBackupSearch->file_hash;
            $backupFound = $this->searchForBackup($fileSize, $fileHash);
            if ($backupFound !== null) {
                try {
                    return File::copyFileToAbstractedPath($backupFound, $newPath);
                } catch (\Exception $e) {
                    throw $e;
                }
            }
        }

        throw $throwable;
    }

    /**
     * @param int $fileSize
     * @param string $fileHash
     * @return string|null
     */
    protected function searchForBackup($fileSize, $fileHash)
    {
        $found = null;
        if ($this->backupPathPrepared === null) {
            return $found;
        }

        $fh = fopen($this->backupPathPrepared, 'r');
        if ($fh === false) {
            return $found;
        }

        while (true) {
            $line = fgets($fh);
            if ($line === false) {
                break;
            }

            if (strpos($line, $fileHash) === false) {
                continue;
            }

            $line = rtrim($line);
            if (filesize($line) !== $fileSize) {
                continue;
            }

            $found = $line;
            break;
        }

        $this->storage()->log('searchForBackup(%d, %s) -> %s', $fileSize, $fileHash, $found);

        fclose($fh);
        return $found;
    }
}
