<?php

namespace Xfrocks\AttachmentStore\Service;

use XF\Db\Schema\Create;
use XF\Service\AbstractService;
use XF\Util\Php;
use Xfrocks\AttachmentStore\Data\Storage;
use Xfrocks\AttachmentStore\Util\EngineOptions;
use Xfrocks\AttachmentStore\XF\Entity\AttachmentData;

class Updater extends AbstractService
{
    const BACKUP_TABLE_NAME_PREFIX = 'bdattachmentstore_updater_';
    const BACKUP_TABLE_DATA_ID = 'data_id';
    const BACKUP_TABLE_OLD_ENGINE = 'old_engine';
    const BACKUP_TABLE_OLD_OPTIONS = 'old_options';
    const BACKUP_TABLE_NEW_ENGINE = 'new_engine';
    const BACKUP_TABLE_NEW_OPTIONS = 'new_options';
    const BACKUP_TABLE_UPDATER_INFO = 'updater_info';

    /**
     * @param AttachmentData $data
     * @param array $rules
     * @return bool
     */
    public function checkRules($data, $rules)
    {
        if (!isset($rules['mode'])) {
            return false;
        }
        $rulesEngine = $rules['mode'];

        $rulesOptions = [];
        if (isset($rules[$rulesEngine])) {
            $rulesOptions = $rules[$rulesEngine];
        }

        return EngineOptions::checkStorageOptionsSame(
            $data->bdattachmentstore_engine,
            $data->bdattachmentstore_options,
            $rulesEngine,
            $rulesOptions
        );
    }

    /**
     * @return string
     */
    public function createBackupTable()
    {
        $tableName = self::BACKUP_TABLE_NAME_PREFIX . \XF::$time;
        $this->app->db()->getSchemaManager()->createTable($tableName, function (Create $table) {
            $table->addColumn(self::BACKUP_TABLE_DATA_ID, 'int')->primaryKey();
            $table->addColumn(self::BACKUP_TABLE_OLD_ENGINE, 'varchar', 100);
            $table->addColumn(self::BACKUP_TABLE_OLD_OPTIONS, 'mediumblob');
            $table->addColumn(self::BACKUP_TABLE_NEW_ENGINE, 'varchar', 100);
            $table->addColumn(self::BACKUP_TABLE_NEW_OPTIONS, 'mediumblob');
            $table->addColumn(self::BACKUP_TABLE_UPDATER_INFO, 'mediumblob');
        });

        return $tableName;
    }

    /**
     * @param AttachmentData $data
     * @param string $backupTable
     * @param bool $verify
     * @return bool
     * @throws \Exception
     */
    public function update($data, $backupTable, $verify)
    {
        /** @var Storage $storage */
        $storage = $this->app->data('Xfrocks\AttachmentStore:Storage');

        $oldEngine = $data->bdattachmentstore_engine;
        $oldOptions = $data->bdattachmentstore_options;

        $newEngine = $storage->getDefaultEngine();
        $newOptions = $storage->getDefaultOptions($newEngine, $data);
        $newOptions['updaterBackupTable'] = $backupTable;
        if (EngineOptions::checkStorageOptionsSame($oldEngine, $oldOptions, $newEngine, $newOptions)) {
            return false;
        }

        $data->bdattachmentstore_engine = $newEngine;
        $data->bdattachmentstore_options = $newOptions;
        if ($verify && !$data->isDataAvailable()) {
            throw new \Exception('Data is unavailable');
        }

        $db = $this->db();
        $db->beginTransaction();

        try {
            $data->save(true, false);

            /** @noinspection PhpParamsInspection */
            $db->insert($backupTable, [
                self::BACKUP_TABLE_DATA_ID => $data->data_id,
                self::BACKUP_TABLE_OLD_ENGINE => $oldEngine,
                self::BACKUP_TABLE_OLD_OPTIONS => Php::safeSerialize($oldOptions),
                self::BACKUP_TABLE_NEW_ENGINE => $data->bdattachmentstore_engine,
                self::BACKUP_TABLE_NEW_OPTIONS => Php::safeSerialize($data->bdattachmentstore_options),
                self::BACKUP_TABLE_UPDATER_INFO => Php::safeSerialize([
                    'user_id' => \XF::visitor()->user_id,
                    'verify' => $verify
                ])
            ]);
        } catch (\Exception $e) {
            $db->rollback();

            throw $e;
        }

        $db->commit();

        return true;
    }
}
