<?php

namespace Xfrocks\AttachmentStore\Util;

use Xfrocks\AttachmentStore\Listener;

class EngineOptions
{
    const CLOUDFRONT_ENABLED = 'enabled';
    const CLOUDFRONT_DOMAIN_NAME = 'domainName';
    const CLOUDFRONT_TTL = 'ttl';
    const CLOUDFRONT_KEY_PAIR_ID = 'keyPairId';
    const CLOUDFRONT_PK = 'pk';
    const CLOUDFRONT_PK_FILE_PATH = 'pkFilePath';

    const FTP_HOST = 'host';
    const FTP_PORT = 'port';
    const FTP_PASSIVE = 'passive';
    const FTP_USERNAME = 'username';
    const FTP_PASSWORD = 'password';
    const FTP_PATH = 'path';
    const FTP_URL = 'url';
    const FTP_CALLBACK = 'callback';

    const GCS_BUCKET = 'bucket';
    const GCS_KEY_FILE_PATH = 'keyFilePath';
    const GCS_PROJECT_ID = 'projectId';
    const GCS_URL = 'url';
    const GCS_CALLBACK = 'callback';

    const S3_REGION = 'region';
    const S3_BUCKET = 'bucket';
    const S3_KEY = 'key';
    const S3_SECRET = 'secret';
    const S3_ENDPOINT = 'endpoint';
    const S3_PRIVATE = 'private';
    const S3_PROTOCOL = 'protocol';
    const S3_PROTOCOL_HTTP = 'http';
    const S3_PROTOCOL_HTTPS = 'https';
    const S3_URL = 'url';
    const S3_CALLBACK = 'callback';
    const S3_SIGNATURE_AWS2 = 'signatureAws2';

    /**
     * @param string|false $firstEngine
     * @param array|null $firstOptions
     * @param string|false $secondEngine
     * @param array|null $secondOptions
     * @return bool
     */
    public static function checkStorageOptionsSame($firstEngine, $firstOptions, $secondEngine, $secondOptions)
    {
        $firstEngine = strval($firstEngine);
        $secondEngine = strval($secondEngine);
        if ($firstEngine !== $secondEngine) {
            return false;
        }

        if (!is_array($firstOptions)) {
            $firstOptions = [];
        }
        if (!is_array($secondOptions)) {
            $secondOptions = [];
        }

        $firstConfig = null;
        $secondConfig = null;
        $preparedConfigs = false;
        switch ($secondEngine) {
            case Listener::MODE_FTP:
                $firstConfig = self::getFtpConfig($firstOptions);
                $secondConfig = self::getFtpConfig($secondOptions);
                $preparedConfigs = true;
                break;
            case Listener::MODE_GCS:
                $firstConfig = self::getGcsConfig($firstOptions);
                $secondConfig = self::getGcsConfig($secondOptions);
                $preparedConfigs = true;
                break;
            case Listener::MODE_S3:
                $firstConfig = self::getS3Config($firstOptions);
                $secondConfig = self::getS3Config($secondOptions);
                $preparedConfigs = true;
                break;
        }

        if ($preparedConfigs) {
            if (!is_array($firstConfig) || !is_array($secondConfig)) {
                return false;
            }

            foreach (array_keys($secondConfig) as $key) {
                if ($secondConfig[$key] === null) {
                    continue;
                }

                if ($firstConfig[$key] !== $secondConfig[$key]) {
                    return false;
                }
            }
        }

        return true;
    }

    /**
     * @param array $options
     * @return array
     */
    public static function getFtpConfig(array &$options)
    {
        if (isset($options[__METHOD__])) {
            return $options[__METHOD__];
        }

        $config = [
            self::FTP_HOST => self::getConfigString($options, self::FTP_HOST),
            self::FTP_PORT => self::getConfigInt($options, self::FTP_PORT),
            self::FTP_PASSIVE => self::getConfigBoolean($options, self::FTP_PASSIVE),
            self::FTP_USERNAME => self::getConfigString($options, self::FTP_USERNAME),
            self::FTP_PASSWORD => self::getConfigString($options, self::FTP_PASSWORD),
            self::FTP_PATH => self::getConfigString($options, self::FTP_PATH),
            self::FTP_URL => self::getConfigString($options, self::FTP_URL),
            self::FTP_CALLBACK => null,
        ];

        if ($config[self::FTP_HOST] !== null) {
            $configuredServer = self::getConfiguredFtpServer($config[self::FTP_HOST]);
            if (is_array($configuredServer)) {
                foreach (array_keys($config) as $key) {
                    if (isset($configuredServer[$key])) {
                        $config[$key] = $configuredServer[$key];
                    }
                }
            }
        }

        $options[__METHOD__] = $config;

        return $config;
    }

    /**
     * @param array $options
     * @return array
     */
    public static function getGcsConfig(array &$options)
    {
        if (isset($options[__METHOD__])) {
            return $options[__METHOD__];
        }

        $config = [
            self::GCS_BUCKET => self::getConfigString($options, self::GCS_BUCKET),
            self::GCS_KEY_FILE_PATH => null,
            self::GCS_PROJECT_ID => null,
            self::GCS_URL => null,
            self::GCS_CALLBACK => null,
        ];

        if ($config[self::GCS_BUCKET] !== null) {
            $configuredBucket = self::getConfiguredGcsBucket($config[self::GCS_BUCKET]);
            if (is_array($configuredBucket)) {
                foreach (array_keys($config) as $key) {
                    if (isset($configuredBucket[$key])) {
                        $config[$key] = $configuredBucket[$key];
                    }
                }
            }
        }

        $options[__METHOD__] = $config;

        return $config;
    }

    /**
     * @param array $options
     * @return array
     */
    public static function getS3Config(array &$options)
    {
        if (isset($options[__METHOD__])) {
            return $options[__METHOD__];
        }

        $config = [
            self::S3_BUCKET => self::getConfigString($options, self::S3_BUCKET),
            self::S3_KEY => self::getConfigString($options, self::S3_KEY),
            self::S3_SECRET => self::getConfigString($options, self::S3_SECRET),
            self::S3_REGION => self::getS3Region($options),
            self::S3_ENDPOINT => null,
            self::S3_PRIVATE => self::getConfigBoolean($options, self::S3_PRIVATE),
            self::S3_PROTOCOL => self::getConfigString($options, self::S3_PROTOCOL),
            self::S3_URL => null,
            self::S3_CALLBACK => null,
            self::S3_SIGNATURE_AWS2 => null,
        ];

        if ($config[self::S3_BUCKET] !== null) {
            $configuredBucket = self::getConfiguredS3Bucket($config[self::S3_BUCKET]);
            if (is_array($configuredBucket)) {
                foreach (array_keys($config) as $key) {
                    if (isset($configuredBucket[$key])) {
                        $config[$key] = $configuredBucket[$key];
                    }
                }
            }
        }

        if ($config[self::S3_SIGNATURE_AWS2] !== null) {
            throw new \InvalidArgumentException(sprintf('%s is no longer supported', self::S3_SIGNATURE_AWS2));
        }

        $options[__METHOD__] = $config;

        return $config;
    }

    /**
     * @param array $options
     * @param string $key
     * @return bool|null
     */
    private static function getConfigBoolean(array $options, $key)
    {
        if (!isset($options[$key]) || $options[$key] === '') {
            return null;
        }

        return $options[$key] === '1';
    }

    /**
     * @param array $options
     * @param string $key
     * @return int|null
     */
    private static function getConfigInt(array $options, $key)
    {
        if (!isset($options[$key]) || $options[$key] === '') {
            return null;
        }

        return intval($options[$key]);
    }

    /**
     * @param array $options
     * @param string $key
     * @return string|null
     */
    private static function getConfigString(array $options, $key)
    {
        if (!isset($options[$key]) || $options[$key] === '') {
            return null;
        }

        return strval($options[$key]);
    }

    /**
     * @param string $host
     * @return array|null
     */
    private static function getConfiguredFtpServer($host)
    {
        static $servers = null;

        if ($servers === null) {
            $servers = \XF::config(Listener::CONFIG_FTP_SERVERS);
            if (!is_array($servers)) {
                $servers = [];
            }
        }

        if (isset($servers[$host])) {
            return $servers[$host];
        } else {
            return null;
        }
    }

    /**
     * @param string $bucket
     * @return array|null
     */
    private static function getConfiguredGcsBucket($bucket)
    {
        static $buckets = null;

        if ($buckets === null) {
            $buckets = \XF::config(Listener::CONFIG_GCS_BUCKETS);
            if (!is_array($buckets)) {
                $buckets = [];
            }
        }

        if (isset($buckets[$bucket])) {
            return $buckets[$bucket];
        } else {
            return null;
        }
    }

    /**
     * @param string $bucket
     * @return array|null
     */
    private static function getConfiguredS3Bucket($bucket)
    {
        static $buckets = null;

        if ($buckets === null) {
            $buckets = \XF::config(Listener::CONFIG_S3_BUCKETS);
            if (!is_array($buckets)) {
                $buckets = [];
            }
        }

        if (isset($buckets[$bucket])) {
            return $buckets[$bucket];
        } else {
            return null;
        }
    }

    /**
     * @param array $options
     * @return string|null
     */
    private static function getS3Region(array $options)
    {
        $regions = S3::getRegions();

        if (isset($options[self::S3_REGION]) &&
            is_string($options[self::S3_REGION]) &&
            isset($regions[$options[self::S3_REGION]])
        ) {
            return $options[self::S3_REGION];
        }

        return null;
    }
}
